/*
 * This is a reverse-engineered driver for mobile WiMAX (802.16e) devices
 * based on GCT Semiconductor GDM7213 & GDM7205 chip.
 * Copyright (�) 2010 Yaroslav Levandovsky <leyarx@gmail.com>
 *
 * Based on  madWiMAX driver writed by Alexander Gordeev
 * Copyright (C) 2008-2009 Alexander Gordeev <lasaine@lvk.cs.msu.su>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>  //typedef
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h> // for mkdir
#include <sys/time.h>
#include <sys/wait.h>

#include <libusb-1.0/libusb.h>

//#include "config.h"
#include "logging.h"
#include "protocol.h"
#include "gctwimax.h"
#include "tap_dev.h"

#include "eap_auth.h"

#include <openssl/sha.h>
#include <openssl/aes.h>
#include "zlib.h"
/* variables for the command-line parameters */

//static int diode_on = 1;
//static int detach_dvd = 0;
//static char *ssid = "@yota.ru";

//struct timeval start_sec, curr_sec; // get link status every second


//static FILE *logfile = NULL;

#define MATCH_BY_LIST		0
#define MATCH_BY_VID_PID	1
#define MATCH_BY_BUS_DEV	2

//static int match_method = MATCH_BY_LIST;

/* for matching by list...
typedef struct usb_device_id_t {
	unsigned short vendorID;
	unsigned short productID;
	unsigned short targetVID;
	unsigned short targetPID;
} usb_device_id_t; */

/* list of all known devices 
static usb_device_id_t wimax_dev_ids[] = {
	{ 0x04e8, 0x6761 },
	{ 0x04e9, 0x6761 },
	{ 0x04e8, 0x6731 },
	{ 0x04e8, 0x6780 },
};*/

/* for other methods of matching... 
static union {
	struct {
		unsigned short vid;
		unsigned short pid;
	};
	struct {
		unsigned int bus;
		unsigned int dev;
	};
} match_params;*/

/* USB-related parameters */
#ifdef GDM7205
	#define IF_MODEM		0	// 0 - для модема комстар	GDM7205
#else
	#define IF_MODEM		1	//1 - для модема фрештел	GDM7213
#endif

#define IF_DVD			0

#define EP_IN			(130 | LIBUSB_ENDPOINT_IN)
#define EP_OUT			(1 | LIBUSB_ENDPOINT_OUT)



/* information collector */
//static struct wimax_dev_status wd_status;

//char *wimax_states[] = {"INIT", "SYNC", "NEGO", "NORMAL", "SLEEP", "IDLE", "HHO", "FBSS", "RESET", "RESERVED", "UNDEFINED", "BE", "NRTPS", "RTPS", "ERTPS", "UGS", "INITIAL_RNG", "BASIC", "PRIMARY", "SECONDARY", "MULTICAST", "NORMAL_MULTICAST", "SLEEP_MULTICAST", "IDLE_MULTICAST", "FRAG_BROADCAST", "BROADCAST", "MANAGEMENT", "TRANSPORT"};

/* libusb stuff */
//static struct libusb_context *ctx = NULL;
//static struct libusb_device_handle *devh = NULL;
//static struct libusb_transfer *req_transfer = NULL;
//static int kernel_driver_active = 0;

//static unsigned char read_buffer[MAX_PACKET_LEN];

//static int tap_fd = -1;
//static char tap_dev[20] = "wimax%d";
//static int tap_if_up = 0;

//static nfds_t nfds;
//static struct pollfd* fds = NULL;

//static int extract_cert(int *id, unsigned char *buf, int *len);

//static int first_nego_flag = 0;
//static int device_disconnected = 0;

#define CHECK_NEGATIVE(x) {if((r = (x)) < 0) return r;}
#define CHECK_DISCONNECTED(x) {if((r = (x)) == LIBUSB_ERROR_NO_DEVICE) { ctx->on_device_disconnect_callback(ctx->user_data_for_callback); return -1;}}

//int error_code = 0;
//int is_dev_found = 0;
//char *conf_dir;

//gct_callback on_device_disconnect;
//void* user_data_callback;

/*
void _init(void) {
	wd_status.rf_state = 1;
}
*/
/*
static struct libusb_device_handle* find_wimax_device(void)
{
	struct libusb_device **devs;
	struct libusb_device *found = NULL;
	struct libusb_device *dev;
	struct libusb_device_handle *handle = NULL;
	int i = 0;
	int r;

	if (libusb_get_device_list(ctx, &devs) < 0)
		return NULL;

	while (!found && (dev = devs[i++]) != NULL) {
		struct libusb_device_descriptor desc;
		unsigned int j = 0;
		unsigned short dev_vid, dev_pid;

		r = libusb_get_device_descriptor(dev, &desc);
		if (r < 0) {
			continue;
		}
		dev_vid = libusb_le16_to_cpu(desc.idVendor);
		dev_pid = libusb_le16_to_cpu(desc.idProduct);
		wmlog_msg(1, "Bus %03d Device %03d: ID %04x:%04x", libusb_get_bus_number(dev), libusb_get_device_address(dev), dev_vid, dev_pid);
		switch (match_method) {
			case MATCH_BY_LIST: {
				for (j = 0; j < sizeof(wimax_dev_ids) / sizeof(usb_device_id_t); j++) {
					if (dev_vid == wimax_dev_ids[j].vendorID && dev_pid == wimax_dev_ids[j].productID) {
						found = dev;
						break;
					}
				}
				break;
			}
			case MATCH_BY_VID_PID: {
				if (dev_vid == match_params.vid && dev_pid == match_params.pid) {
					found = dev;
				}
				break;
			}
			case MATCH_BY_BUS_DEV: {
				if (libusb_get_bus_number(dev) == match_params.bus && libusb_get_device_address(dev) == match_params.dev) {
					found = dev;
				}
				break;
			}
		}
	}

	if (found) {
		r = libusb_open(found, &handle);
		if (r < 0)
			handle = NULL;
	}

	libusb_free_device_list(devs, 1);
	return handle;
}
*/

////////////////////////////////////////////
////////////////////////////////////////////////
///////////////////////////////////////////////

gctwimax_wimax_t *gctwimax_wimax_new() {
	gctwimax_wimax_t *handle = (gctwimax_wimax_t *) malloc(sizeof(gctwimax_wimax_t));
	handle->libusb_ctx = NULL;
	handle->libusb_devh = NULL;
	handle->device_disconnected = 0;
	handle->error_code = 0;
	handle->fds = NULL;
	handle->kernel_driver_active = 0;
	handle->libusb_req_transfer = NULL;
	strcpy(handle->tap_dev, "wimax%d");
	handle->tap_fd = -1;
	handle->tap_if_up = 0;
	handle->connection_status.rf_state = 1;
	return handle;
}

/////////////////////////////////////////////
//////////////////////////////////////////////
/////////////////////////////////////////////

// ***Edited by fanboy*** 
static struct libusb_device_handle* find_wimax_device(gctwimax_wimax_t *ctx)
{
	struct libusb_device_handle *handle = NULL;
	int r;
	//Switch modem the same as in usb_modeswitch
	handle = libusb_open_device_with_vid_pid(NULL, 0x1076, 0x7f40);
	if (handle != 0 ){
		
		if (libusb_kernel_driver_active(handle, IF_MODEM) == 1)
		{
			r = libusb_detach_kernel_driver(handle, IF_MODEM);
			if (r < 0){
				wmlog_msg(1, "Kernel driver detaching (error %d)\n", r);
			} else {
				wmlog_msg(1, "Kernel driver deteched!\n");
			}
		}	
	
		r = libusb_claim_interface(handle,IF_MODEM);

		if (r < 0) {
			wmlog_msg(1, "Claim Interface problems (error %d)\n", r);
		}
		else
		{
			wmlog_msg(1, "Innterface claimed\n");	

			r = libusb_control_transfer(handle, 0xa1, 0xa0, 0, IF_MODEM, ctx->read_buffer, 1, 1000);
			wmlog_msg(1, "Sending Control message (result %d - %s)\n", r, r ? "bad" : "ok");

			libusb_release_interface(handle, IF_MODEM);

			libusb_close(handle);
		}
	}

	int retry = 0;
	do
	{
		sleep(1); // Wait while device switching
		handle = libusb_open_device_with_vid_pid(NULL, 0x1076, 0x7f00);
	}
	while (retry++ < 5 && !handle);

	if (handle) wmlog_msg(2, "Device switched after %d retries.\n", retry);
	else  wmlog_msg(1, "Device not switched after %d retries.\n", retry);

	return handle;
}

static int set_data(gctwimax_wimax_t *ctx, unsigned char* data, int size)
{
	int r;
	int transferred;

	wmlog_dumphexasc(3, data, size, "Bulk write:");

	r = libusb_bulk_transfer(ctx->libusb_devh, EP_OUT, data, size, &transferred, 0);
	if (r < 0) {
		wmlog_msg(1, "bulk write error %d", r);
		if (r == LIBUSB_ERROR_NO_DEVICE) {
			ctx->on_device_disconnect_callback(ctx->user_data_for_callback);
			ctx->error_code = 4;
			return -1;
		}
		return r;
	}
	if (transferred < size) {
		wmlog_msg(1, "short write (%d)", r);
		return -1;
	}
	return r;
}

void eap_server_rx(gctwimax_wimax_t *ctx, unsigned char *data, int data_len) // Sent data from peer to server
{
	unsigned char req_data[MAX_PACKET_LEN];
	int len;
	
	len = fill_eap_server_rx_req(req_data, data, data_len);
	set_data(ctx, req_data, len);
}

void eap_key(gctwimax_wimax_t *ctx, unsigned char *data, int data_len)
{
	unsigned char req_data[MAX_PACKET_LEN];
	int len;

	len = fill_eap_key_req(req_data, data, data_len);
	set_data(ctx, req_data, len);
}

static void cb_req(struct libusb_transfer *transfer)
{
	gctwimax_wimax_t *ctx = (gctwimax_wimax_t*) transfer->user_data;

	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		wmlog_msg(1, "async bulk read error %d", transfer->status);
		if (transfer->status == LIBUSB_TRANSFER_NO_DEVICE) {
			ctx->device_disconnected = 1;
			return;
		}
	} else {
		wmlog_dumphexasc(3, transfer->buffer, transfer->actual_length, "Async read:");
		process_response(ctx, transfer->buffer, transfer->actual_length);
	}
	if (libusb_submit_transfer(ctx->libusb_req_transfer) < 0) {
		wmlog_msg(1, "async read transfer sumbit failed");
	}
}

/* get link_status *//*
int get_link_status()
{
	return wd_status.link_status;
}*/

/* set close-on-exec flag on the file descriptor */
int set_coe(int fd)
{
	int flags;

	flags = fcntl(fd, F_GETFD);
	if (flags == -1)
	{
		wmlog_msg(1, "failed to set close-on-exec flag on fd %d", fd);
		return -1;
	}
	flags |= FD_CLOEXEC;
	if (fcntl(fd, F_SETFD, flags) == -1)
	{
		wmlog_msg(1, "failed to set close-on-exec flag on fd %d", fd);
		return -1;
	}

	return 0;
}

/* set link_status *//*
void set_link_status(int link_status)
{
	wd_status.info_updated |= WDS_LINK_STATUS;

	if (wd_status.link_status == link_status) return;

	if (wd_status.link_status < 2 && link_status == 2) {
		if_up();
	}
	if (wd_status.link_status == 2 && link_status < 2) {
		if_down();
	}
	if (link_status == 1) {
		first_nego_flag = 1;
	}

	wd_status.link_status = link_status;
}*/

/* get state *//*
int get_state()
{
	return wd_status.state;
}*/

/* set state *//*
void set_state(int state)
{
	wd_status.state = state;
	wd_status.info_updated |= WDS_STATE;
	if (state >= 1 && state <= 3 && wd_status.link_status != (state - 1)) {
		set_link_status(state - 1);
	}
}*/

static int alloc_transfers(gctwimax_wimax_t *ctx)
{
	ctx->libusb_req_transfer = libusb_alloc_transfer(0);
	if (!ctx->libusb_req_transfer)
		return -ENOMEM;

	libusb_fill_bulk_transfer(ctx->libusb_req_transfer, ctx->libusb_devh, EP_IN, ctx->read_buffer,
		sizeof(ctx->read_buffer), cb_req, ctx, 0);

	return 0;
}

int write_netif(gctwimax_wimax_t *ctx,const void *buf, int count)
{
	return tap_write(ctx->tap_fd, buf, count);
}

static int read_tap(gctwimax_wimax_t *ctx)
{
	unsigned char buf[MAX_PACKET_LEN];
	int hlen = get_header_len();
	int r;
	int len;

	r = tap_read(ctx->tap_fd, buf + hlen, MAX_PACKET_LEN - hlen);

	if (r < 0)
	{
		wmlog_msg(1, "Error while reading from TAP interface");
		return r;
	}

	if (r == 0)
	{
		return 0;
	}

	len = fill_data_packet_header(buf, r);
	wmlog_dumphexasc(4, buf, len, "Outgoing packet:");
	r = set_data(ctx, buf, len);

	return r;
}

static int process_events_once(gctwimax_wimax_t *ctx, int timeout)
{
	struct timeval tv = {0, 0};
	int r;
	int libusb_delay;
	int delay;
	unsigned int i;
	char process_libusb = 0;

	r = libusb_get_next_timeout(ctx->libusb_ctx, &tv);
	if (r == 1 && tv.tv_sec == 0 && tv.tv_usec == 0)
	{
		r = libusb_handle_events_timeout(ctx->libusb_ctx, &tv);
	}

	delay = libusb_delay = tv.tv_sec * 1000 + tv.tv_usec;
	if (delay <= 0 || delay > timeout)
	{
		delay = timeout;
	}

	CHECK_NEGATIVE(poll(ctx->fds, ctx->nfds, delay));

	process_libusb = (r == 0 && delay == libusb_delay);

	for (i = 0; i < ctx->nfds; ++i)
	{
		if (ctx->fds[i].fd == ctx->tap_fd) {
			if (ctx->fds[i].revents)
			{
				CHECK_NEGATIVE(read_tap(ctx));
			}
			continue;
		}
		process_libusb |= ctx->fds[i].revents;
	}

	if (process_libusb)
	{
		struct timeval tv = {.tv_sec = 0, .tv_usec = 0};
		CHECK_NEGATIVE(libusb_handle_events_timeout(ctx->libusb_ctx, &tv));
	}

	return 0;
}

/* handle events until timeout is reached or all of the events in event_mask happen */
static int process_events_by_mask(gctwimax_wimax_t *ctx, int timeout, unsigned int event_mask)
{
	struct timeval start, curr;
	int r;
	int delay = timeout;

	CHECK_NEGATIVE(gettimeofday(&start, NULL));

	ctx->connection_status.info_updated &= ~event_mask;
	
	while ((event_mask == 0 || (ctx->connection_status.info_updated & event_mask) != event_mask) && delay >= 0) {
		long a;
		
		CHECK_NEGATIVE(process_events_once(ctx, delay));

		if (ctx->device_disconnected) {
			ctx->on_device_disconnect_callback(ctx->user_data_for_callback);
			return -1;
		}

		CHECK_NEGATIVE(gettimeofday(&curr, NULL));

		a = (curr.tv_sec - start.tv_sec) * 1000 + (curr.tv_usec - start.tv_usec) / 1000;
		delay = timeout - a;
	}

	ctx->connection_status.info_updated &= ~event_mask;

	return (delay > 0) ? delay : 0;
}

int alloc_fds(gctwimax_wimax_t *ctx)
{
	int i;
	const struct libusb_pollfd **usb_fds = libusb_get_pollfds(ctx->libusb_ctx);

	if (!usb_fds)
	{
		return -1;
	}

	ctx->nfds = 0;
	while (usb_fds[ctx->nfds])
	{
		ctx->nfds++;
	}
	if (ctx->tap_fd != -1) {
		ctx->nfds++;
	}

	if(ctx->fds != NULL) {
		free(ctx->fds);
	}

	ctx->fds = (struct pollfd*)calloc(ctx->nfds, sizeof(struct pollfd));
	for (i = 0; usb_fds[i]; ++i)
	{
		ctx->fds[i].fd = usb_fds[i]->fd;
		ctx->fds[i].events = usb_fds[i]->events;
		set_coe(usb_fds[i]->fd);
	}
	if (ctx->tap_fd != -1) {
		ctx->fds[i].fd = ctx->tap_fd;
		ctx->fds[i].events = POLLIN;
		ctx->fds[i].revents = 0;
	}

	free(usb_fds);

	return 0;
}

void cb_add_pollfd(int fd, short events, void *user_data)
{
	alloc_fds(user_data);
}

void cb_remove_pollfd(int fd, void *user_data)
{
	alloc_fds(user_data);
}

void gctwimax_cb_add_pollfd(gctwimax_wimax_t *ctx) {
	cb_add_pollfd(ctx->tap_fd, POLLIN, ctx);
}
/*
void gctwimax_remove_pollfd(void) {
	cb_remove_pollfd(tap_fd, NULL);
}
*/
void AES_cbc_decrypt(uint8_t *shakey, uint8_t *inbuf, uint8_t *outbuf)
{
	unsigned char iv[16] = {0x43,0x6c,0x61,0x72,0x6b,0x4a,0x4a,0x61,0x6e,0x67,0x00,0x00,0x00,0x00,0x00,0x00}; //ClarkJJang

	AES_KEY aeskey;	

	AES_set_decrypt_key(shakey, 24*8, &aeskey); //192

	AES_cbc_encrypt(inbuf, outbuf, 16, &aeskey, iv, AES_DECRYPT);
}

static int extract_cert(gctwimax_wimax_t *ctx, int id, unsigned char *buf, int len, char *path)
{
	int r, i;
	int dec_buf_len, gz_len;
	char gz_path[0x100], pem_path[0x100];
	uint8_t key_sha[0x18];
	uint8_t dec_buf[len], gz_buf[MAX_PACKET_LEN];
	FILE *fp;
	voidp gz;
	
	memset(key_sha,0x00,sizeof(key_sha));
	memset(dec_buf,0x00,sizeof(dec_buf));
	
	r = mkdir(path,S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (r)
	{
		if (errno != 17)
			return errno;
	}
	

	SHA1(ctx->device_info.mac, 6, key_sha);

	for (i = 0; i < (len / 16); i++)
		AES_cbc_decrypt(key_sha, buf + 16 * i, dec_buf + 16 * i);	
	dec_buf_len = (dec_buf[0] << 24) + (dec_buf[1] << 16) + (dec_buf[2] << 8) + dec_buf[3];
	wmlog_msg(2, "Decrypted cert %d size is %d bytes.", id, dec_buf_len);
	
	sprintf(gz_path,"%s/0%d.gz", path, id);
	fp=fopen((char *)gz_path, "wb");
	if (fp == NULL)
		return errno;
	fwrite(dec_buf + 4, 1, dec_buf_len, fp);
	fclose(fp);

	gz = gzopen(gz_path, "rb"); 
	if (gz == NULL)
		return errno;
	gz_len = gzread(gz, gz_buf, sizeof(gz_buf));
	gzclose(gz);
	remove(gz_path);
	
	wmlog_msg(2, "Extracted cert %d size is %d bytes.", id, gz_len);

	if(id == 1)
	{
		sprintf(pem_path, "%s/client_cert.pem", path);

		fp=fopen((char *)pem_path, "wb");
		if (fp == NULL)
			return errno;
		fwrite(gz_buf,1,gz_len,fp);
		fclose(fp);			
	}
	else
	{
		sprintf(pem_path, "%s/root_ca.pem", path);
		
		fp=fopen((char *)pem_path, "ab");
		if (fp == NULL)
			return errno;
		fwrite(gz_buf,1,gz_len,fp);
		fclose(fp);
	}
	
	return 0;
}

void gctwimax_set_log_full(int verbose, int logger, FILE *logfile) {
	gctwimax_set_log(logger, logfile);
	set_wmlog_level(verbose);
}

void gctwimax_set_log(int logger, FILE *logfile) {
	if (logger == WMLOGGER_SYSLOG) {
		set_wmlogger("gctwimax", WMLOGGER_SYSLOG, NULL);
	} else {
		gctwimax_set_log_file(logfile);
	}
}

void gctwimax_set_log_file(FILE *logfile) {
	if (logfile != NULL) {
		set_wmlogger("gctwimax", WMLOGGER_FILE, logfile);
	} else {
		set_wmlogger("gctwimax", WMLOGGER_FILE, stderr);
	}
}

void gctwimax_set_log_level(int verbose) {
	set_wmlog_level(verbose);
}

void gctwimax_inc_log_level(void) {
	inc_wmlog_level();
}

int gctwimax_search_network(gctwimax_wimax_t *ctx) {
	unsigned char req_data[MAX_PACKET_LEN];
	int len;

	len = fill_auth_on_req(req_data);
	set_data(ctx, req_data, len);

	wmlog_msg(0, "Search network...");
	len = fill_find_network_req(req_data, &ctx->connection_status);
	set_data(ctx, req_data, len);

	while (process_events_by_mask(ctx, 60000, WDS_LINK_STATUS) < 0) {;}

	if (ctx->connection_status.link_status == 0) {
		wmlog_msg(0, "Network not found.");
		return 0;
	} else {
		wmlog_msg(0, "Network found.");
		return 1;
	}
	return 0;
}

int gctwimax_connected_to_network(gctwimax_wimax_t *ctx) {
	unsigned char req_data[MAX_PACKET_LEN];
	int len;

	ctx->connection_status.link_status = 0;

	len = fill_connect_req(req_data, &ctx->connection_status);
	set_data(ctx, req_data, len);

	while (process_events_by_mask(ctx, 100000, WDS_LINK_STATUS) < 0) {;}

	if (ctx->connection_status.link_status == 0) {
		wmlog_msg(0, "Connection error.");
		return 0;
	} else {
		wmlog_msg(0, "Connected to network.");
		return 1;
	}
	return 0;
}

int gctwimax_start_authentication(gctwimax_wimax_t *ctx) {
	int r;
	r = 1;
	process_events_by_mask(ctx, 5000, WDS_AUTH_STATE);
	wmlog_msg(2, "Auth_state: %06x",ctx->connection_status.auth_state);

	if (ctx->connection_status.auth_state == 0x0001ff ||
		ctx->connection_status.auth_state == 0x0100ff ||
		ctx->connection_status.auth_state == 0x0101ff ||
		ctx->connection_status.auth_state == 0x0200ff ||
		ctx->connection_status.auth_state == 0x0300ff ||
		ctx->connection_status.auth_state == 0x0301ff ||
		ctx->connection_status.auth_state == 0x0400ff ||
		ctx->connection_status.auth_state == 0x0401ff){
		process_events_by_mask(ctx, 3000, WDS_LINK_STATUS);
		//process_events_by_mask(3000, WDS_LINK_STATUS);
		return 1;
	}

	if (ctx->connection_status.auth_state == 0x020000){
		wmlog_msg(0, "Start Authentication.");

		if (eap_peer_init(&ctx->config)< 0)
			wmlog_msg(0, "EAP Peer init error.");

			ctx->connection_status.info_updated = WDS_NONE;
			r = 1;

		while( r > 0)
		{
			while (process_events_by_mask(ctx, 1500, WDS_EAP) < 0) {;}
			//process_events_by_mask(1500, WDS_EAP);
			r = eap_peer_step(ctx);
			//wmlog_msg(0, "\nr = %d\n",r);
		}
		eap_peer_deinit(&ctx->config);
		return 2;
	}

	if (ctx->connection_status.auth_state == 0x0201ff){
		if (ctx->connection_status.auth_info == 0){
			wmlog_msg(0, "Authentication Failed. Renewing Authentication.");
			return 3;
		} else {
			wmlog_msg(0, "Authentication Failed.");
			return 4;
		}
	}

	if (ctx->connection_status.auth_state == 0x020100){
		wmlog_msg(0, "Authentication Succeed.");
		return 5;

		//process_events_by_mask(1000, WDS_LINK_STATUS);

		//process_events_by_mask(500, WDS_OTHER);

	}

	if (ctx->connection_status.auth_state == 0x040100){
		process_events_by_mask(ctx, 2000, WDS_LINK_STATUS);
		process_events_by_mask(ctx, 2000, WDS_LINK_STATUS);
		return 6;
	}

	return 0;
}

int gctwimax_init(gctwimax_wimax_t *ctx) {
	unsigned char req_data[MAX_PACKET_LEN], cert_buf[MAX_PACKET_LEN];
	char path[0x100];
	int len, cert_buf_len, cert_id, cert_len;
	int r = 1;

	r = libusb_init(&ctx->libusb_ctx);
	if (r < 0) {
		wmlog_msg(0, "failed to initialise libusb");
		ctx->error_code = 1;
		return -1;
	}

	ctx->libusb_devh = find_wimax_device(ctx);
	if (ctx->libusb_devh == NULL) {
		wmlog_msg(0, "Could not find/open device");
		ctx->error_code = 2;
		return -1;
	}

	wmlog_msg(0, "Device found");

	if (libusb_kernel_driver_active(ctx->libusb_devh, IF_DVD) == 1) {
		ctx->kernel_driver_active = 1;
		r = libusb_detach_kernel_driver(ctx->libusb_devh, IF_DVD);
		if (r < 0) {
			wmlog_msg(0, "kernel driver detach error %d", r);
		} else {
			wmlog_msg(0, "detached modem kernel driver");
		}
	}

	r = libusb_claim_interface(ctx->libusb_devh, 0);
	if (r < 0) {
		wmlog_msg(0, "Claim usb interface error %d", r);
		ctx->error_code = 3;
		return -1;
	}
	wmlog_msg(0, "Claimed interface");

	alloc_fds(ctx);
	libusb_set_pollfd_notifiers(ctx->libusb_ctx, cb_add_pollfd, cb_remove_pollfd, ctx);

	alloc_transfers(ctx);

	wmlog_msg(2, "Continuous async read start...");
	CHECK_DISCONNECTED(libusb_submit_transfer(ctx->libusb_req_transfer));

	len = fill_debug_req(req_data);  //mode normal or test
	set_data(ctx, req_data, len);


	len = fill_string_info_req(req_data);
	set_data(ctx, req_data, len);

	process_events_by_mask(ctx, 500, WDS_CHIP | WDS_FIRMWARE);

	wmlog_msg(1, "Chip info: %s", ctx->device_info.chip);
	wmlog_msg(1, "Firmware info: %s", ctx->device_info.firmware);

	len = fill_mac_req(req_data);
	set_data(ctx, req_data, len);

	process_events_by_mask(ctx, 500, WDS_MAC);

	wmlog_msg(1, "MAC: %02x:%02x:%02x:%02x:%02x:%02x", ctx->device_info.mac[0], ctx->device_info.mac[1], ctx->device_info.mac[2], \
								ctx->device_info.mac[3], ctx->device_info.mac[4], ctx->device_info.mac[5]);
	sprintf(path,"%s/%02x%02x%02x%02x%02x%02x", ctx->config.conf_dir, ctx->device_info.mac[0], ctx->device_info.mac[1], \
								ctx->device_info.mac[2], ctx->device_info.mac[3], ctx->device_info.mac[4], ctx->device_info.mac[5]);



	ctx->config.eap_ca_path = (char *)malloc(strlen(path)+13);
	sprintf(ctx->config.eap_ca_path,"%s/root_ca.pem",path);

	ctx->config.eap_client_cert = (char *)malloc(strlen(path)+17);
	sprintf(ctx->config.eap_client_cert,"%s/client_cert.pem",path);


	len = fill_rf_on_req(req_data);
	set_data(ctx, req_data, len);

	process_events_by_mask(ctx, 20000, WDS_RF_STATE);

	wmlog_msg(2, "RF ON.");

	memset(ctx->device_info.cert,0xff,6);
	cert_buf_len = 0;


	//wd_status.cert[1] <= 0x06
	while(1)
	{
		cert_id = ctx->device_info.cert[1];

		if(ctx->device_info.cert[1] == 0x06 && ctx->device_info.cert[2] == 0xff)
			break;
		len = fill_get_cert_req(req_data, &ctx->device_info);
		if(len == 0)
			break;
		set_data(ctx, req_data, len);
		process_events_by_mask(ctx, 2000, WDS_CERT);

		cert_len = (ctx->device_info.cert_buf[0] << 8) + ctx->device_info.cert_buf[1];


		if ((cert_id == ctx->device_info.cert[1] && ctx->device_info.cert[2] != 0xff) || (cert_id != ctx->device_info.cert[1] && cert_len > 0))
		{
			memcpy(cert_buf+cert_buf_len,ctx->device_info.cert_buf+2,cert_len);
			cert_buf_len += cert_len;
		}
		else if (cert_id == ctx->device_info.cert[1] && cert_buf_len > 0)
		{
			if (!(cert_buf_len % 16))
			{
				r = extract_cert(ctx, cert_id, cert_buf, cert_buf_len, path);
				if(r)
					wmlog_msg(1, "Error: \"%s\" while extracting cert %d.",strerror(r),cert_id);
				else
				{
					wmlog_msg(1, "Cert %d extracted successfully",cert_id);
				}
			}
			else
				wmlog_msg(1, "Bad cert size %d bytes.",cert_buf_len);

			memset(ctx->device_info.cert_buf,0x00,sizeof(ctx->device_info.cert_buf));
			cert_buf_len = 0;
		}
	}
	return 0;

}

int gctwimax_keep_connection(gctwimax_wimax_t *ctx) {
	unsigned char req_data[MAX_PACKET_LEN];
	int len;

	len = fill_connection_params_req(req_data);
	set_data(ctx, req_data, len);

	process_events_by_mask(ctx, 500, WDS_RSSI1 | WDS_CINR1 | WDS_RSSI2 | WDS_CINR2 | WDS_TXPWR | WDS_FREQ);

	wmlog_msg(1, "RSSI1: %d   CINR1: %d   TX Power: %d   Frequency: %d",
				ctx->connection_status.rssi1, ctx->connection_status.cinr1, ctx->connection_status.txpwr, ctx->connection_status.freq);
	wmlog_msg(1, "RSSI2: %d   CINR2: %d",
					ctx->connection_status.rssi2, ctx->connection_status.cinr2);

	process_events_by_mask(ctx, 5000, WDS_LINK_STATUS);
	return 0;
}

int gctwimax_init_tap(gctwimax_wimax_t *ctx) {
	ctx->tap_fd = tap_open(ctx->tap_dev);
	if (ctx->tap_fd < 0) {
		wmlog_msg(0, "failed to allocate tap interface");
		wmlog_msg(0,
				"You should have TUN/TAP driver compiled in the kernel or as a kernel module.\n"
				"If 'modprobe tun' doesn't help then recompile your kernel.");
		return -1;
	}
	tap_set_hwaddr(ctx->tap_fd, ctx->tap_dev, ctx->device_info.mac);
	tap_set_mtu(ctx->tap_fd, ctx->tap_dev, 1386);
	set_coe(ctx->tap_fd);
	wmlog_msg(0, "Allocated tap interface: %s", ctx->tap_dev);
	return 0;
}

int gctwimax_up_tap(gctwimax_wimax_t *ctx) {
	tap_bring_up(ctx->tap_fd, ctx->tap_dev);
	ctx->tap_if_up = 1;
	return 0;
}

int gctwimax_deinit_tap(gctwimax_wimax_t *ctx) {
	tap_close(ctx->tap_fd, ctx->tap_dev);
	return 0;
}

int gctwimax_down_tap(gctwimax_wimax_t *ctx) {
	if (!ctx->tap_if_up) return 0;
	ctx->tap_if_up = 0;
	tap_bring_down(ctx->tap_fd, ctx->tap_dev);
	return 0;
}

int gctwimax_release_resources(gctwimax_wimax_t *ctx) {
	//Rewrite this part!!!!!!!
	if(ctx->connection_status.rf_state == 0){
		unsigned char req_data[MAX_PACKET_LEN];
		int len;
		wmlog_msg(2, "RF OFF.");
		len = fill_rf_off_req(req_data);
		set_data(ctx, req_data, len);
		process_events_by_mask(ctx, 2000, WDS_RF_STATE);
	}

	remove(ctx->config.eap_ca_path); //Delete this
	remove(ctx->config.eap_client_cert);
	
	if(ctx->libusb_ctx != NULL) {
		if(ctx->libusb_req_transfer != NULL) {
			libusb_cancel_transfer(ctx->libusb_req_transfer);
			libusb_free_transfer(ctx->libusb_req_transfer);
		}
		libusb_set_pollfd_notifiers(ctx->libusb_ctx, NULL, NULL, NULL);
		if(ctx->fds != NULL) {
			free(ctx->fds);
		}
		if(ctx->libusb_devh != NULL) {
			libusb_release_interface(ctx->libusb_devh, 0);
			if (ctx->kernel_driver_active)
				libusb_attach_kernel_driver(ctx->libusb_devh, 0);
			libusb_unlock_events(ctx->libusb_ctx);
			libusb_close(ctx->libusb_devh);
		}
		libusb_exit(ctx->libusb_ctx);
	}

	return 1;
}

void gctwimax_wimax_free(gctwimax_wimax_t *ctx) {
	gctwimax_release_resources(ctx);
	free(ctx);
}

const char *gctwimax_get_version(void) {
	return GCTWIMAX_VERSION;
}
