/*
 * This is a reverse-engineered driver for mobile WiMAX (802.16e) devices
 * based on GCT Semiconductor GDM7213 & GDM7205 chip.
 * Copyright (�) 2010 Yaroslav Levandovsky <leyarx@gmail.com>
 *
 * Based on  madWiMAX driver writed by Alexander Gordeev
 * Copyright (C) 2008-2009 Alexander Gordeev <lasaine@lvk.cs.msu.su>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <poll.h>

#ifndef _WIMAX_H
#define _WIMAX_H

#ifndef GCTWIMAX_VERSION
#define GCTWIMAX_VERSION "0.0.2"
#endif

#define WDS_NONE	0x000
#define WDS_CHIP	0x001
#define WDS_FIRMWARE	0x002
#define WDS_MAC		0x004
#define WDS_LINK_STATUS	0x008
#define WDS_RSSI1	0x010
#define WDS_RSSI2	0x011
#define WDS_CINR1	0x020
#define WDS_CINR2	0x021
#define WDS_BSID	0x040
#define WDS_TXPWR	0x080
#define WDS_FREQ	0x100
#define WDS_STATE	0x200
#define WDS_PROTO_FLAGS	0x400
#define WDS_OTHER	0x800
#define WDS_ANY		0xfff
#define WDS_NSPID	0xd20 // *
//#define WDS_NETNAME	0xd40 // *
#define WDS_NAPID	0xd60 // *
#define WDS_CERT	0xd80 //*
#define WDS_AUTH_STATE	0xa20//*
#define WDS_RF_STATE	0xa00 // *
#define WDS_EAP		0xf00 //EAP

#define MAX_PACKET_LEN		0x4000


//on device disconnect callback function type
typedef void (*gctwimax_callback) (void* user_data);

typedef struct _gctwimax_connection_t {
	unsigned int info_updated;
	unsigned char proto_flags;
	int link_status;
	short rssi1;
	short rssi2;
	short cinr1;
	short cinr2;
	unsigned char bsid[6];
	unsigned short txpwr;
	unsigned int freq;
	unsigned int nspid;
	unsigned int napid;
	unsigned int auth_state; //*
	int auth_info;
	//char netname[0x40]; //*
	int rf_state; // RF-reciever 0=On 1=OFF
	int state;
} gctwimax_connection_t;

typedef struct _gctwimax_device_t {
	unsigned char cert[6];
	unsigned char cert_buf[1026];
	char chip[0x40];
	char firmware[0x40];
	unsigned char mac[6];
} gctwimax_device_t;

typedef struct _gctwimax_config_t {
	//eap config
	char *eap_login;
	char *eap_pass;
	char *eap_outer_nai;
	char *eap_client_cert;
	char *eap_ca_path;
	int eap_type;
	//other config
	char *conf_dir;
} gctwimax_config_t;

typedef struct _gctwimax_wimax_t {
	struct libusb_context *libusb_ctx;
	struct libusb_device_handle *libusb_devh;
	struct libusb_transfer *libusb_req_transfer;
	int kernel_driver_active;
	unsigned char read_buffer[MAX_PACKET_LEN];
	int tap_fd;
	char tap_dev[20];
	int tap_if_up;
	nfds_t nfds;
	struct pollfd* fds;
	int device_disconnected;
	int error_code;
	gctwimax_callback on_device_disconnect_callback;
	void* user_data_for_callback;
	gctwimax_connection_t connection_status;
	gctwimax_device_t device_info;
	gctwimax_config_t config;
} gctwimax_wimax_t;

/* get/set link_status *//*
int get_link_status();
void set_link_status(int link_status);
*/
/* get/set state *//*
int get_state();
void set_state(int state);
*/

/*send data to eap*/
void eap_server_rx(gctwimax_wimax_t *ctx, unsigned char *data, int data_len);
/* get eap key*/
void eap_key(gctwimax_wimax_t *ctx, unsigned char *data, int data_len);
/* write packet to the network interface */
int write_netif(gctwimax_wimax_t *ctx, const void *buf, int count);


void gctwimax_cb_add_pollfd(gctwimax_wimax_t *ctx);
/*
void gctwimax_cb_remove_pollfd(void);
*/

gctwimax_wimax_t* gctwimax_wimax_new(void);
void gctwimax_wimax_free(gctwimax_wimax_t *ctx);

// Log option functions
void gctwimax_set_log(int logger, FILE *logfile);
void gctwimax_set_log_full(int verbose, int logger, FILE *logfile);
void gctwimax_set_log_file(FILE *logfile);
void gctwimax_set_log_level(int verbose);
void gctwimax_inc_log_level(void);

//Network functions
int gctwimax_search_network(gctwimax_wimax_t *ctx);
int gctwimax_connected_to_network(gctwimax_wimax_t *ctx);
int gctwimax_start_authentication(gctwimax_wimax_t *ctx);
int gctwimax_init(gctwimax_wimax_t *ctx);
int gctwimax_init_tap(gctwimax_wimax_t *ctx);
int gctwimax_up_tap(gctwimax_wimax_t *ctx);
int gctwimax_deinit_tap(gctwimax_wimax_t *ctx);
int gctwimax_down_tap(gctwimax_wimax_t *ctx);
int gctwimax_keep_connection(gctwimax_wimax_t *ctx);
int gctwimax_release_resources(gctwimax_wimax_t *ctx);
const char *gctwimax_get_version(void);
#endif // _WIMAX_H

